use game::Game;
use evaluate::PosValue;
use std::collections::{HashMap};
use std::hash::{BuildHasher, Hasher, Hash};
use log::info;
use zobrist;

#[derive(Clone)]
pub enum EntryType {
    Value,
    LowerBound,
    UpperBound,
}

#[derive(Clone)]
pub struct CacheEntry {
    pub entry_type: EntryType,
    pub depth: i32,
    pub value: PosValue
}

impl CacheEntry {
    pub fn new_value(depth: i32, value: PosValue) -> Self {
        CacheEntry {
            entry_type: EntryType::Value,
            depth,
            value
        }
    }

    pub fn new_beta(depth: i32, beta: PosValue) -> Self {
        CacheEntry {
            entry_type: EntryType::UpperBound,
            depth,
            value: beta
        }
    }

    pub fn new_alpha(depth: i32, alpha: PosValue) -> Self {
        CacheEntry {
            entry_type: EntryType::LowerBound,
            depth,
            value: alpha
        }
    }
}

pub struct Cache {
    hashmap: HashMap<zobrist::Hash, CacheEntry>,
    max_capacity: usize
}

impl Cache {
    pub fn new() -> Self {
        Cache {
            hashmap: HashMap::new(),
            max_capacity: 10000000
        }
    }

    pub fn lookup(&self, game_pos: &Game) -> Option<CacheEntry> {
        self.hashmap.get(&game_pos.zobrist.clone().unwrap().1).map(|x| x.clone())
    }
    
    pub fn cache(&mut self, game_pos: &Game, ce: CacheEntry) {
        if self.hashmap.len() > self.max_capacity {
            let first_key = self.hashmap.keys().next().unwrap().clone();
            self.hashmap.remove(&first_key);
        }
        self.hashmap.insert(game_pos.zobrist.clone().unwrap().1, ce);
        if self.hashmap.len() % (1024 * 32) == 0 {
            info!("hash contains {} items", self.hashmap.len());
        }
    }
}